{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Question 72 - LC 106. Construct Binary Tree from Inorder and Postorder Traversal\n",
    "\n",
    "Given two integer arrays `inorder` and `postorder` where `inorder` is the inorder traversal of a binary tree and `postorder` is the postorder traversal of the same tree, construct and return _the binary tree_.\n",
    "\n",
    "**Example 1:**\n",
    "\n",
    "    Input: inorder = [9,3,15,20,7], postorder = [9,15,7,20,3]\n",
    "    Output: [3,9,20,null,null,15,7]\n",
    "\n",
    "**Example 2:**\n",
    "\n",
    "    Input: inorder = [-1], postorder = [-1]\n",
    "    Output: [-1]\n",
    "\n",
    "**Constraints:**\n",
    "\n",
    "- `1 <= inorder.length <= 3000`\n",
    "- `postorder.length == inorder.length`\n",
    "- `-3000 <= inorder[i], postorder[i] <= 3000`\n",
    "- `inorder` and `postorder` consist of unique values.\n",
    "- Each value of `postorder` also appears in `inorder`.\n",
    "- `inorder` is guaranteed to be the inorder traversal of the tree.\n",
    "- `postorder` is guaranteed to be the postorder traversal of the tree.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "from __future__ import annotations\n",
    "from typing import Optional\n",
    "from collections import deque"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Definition for a binary tree node.\n",
    "class TreeNode:\n",
    "    def __init__(self, val=0, left=None, right=None):\n",
    "        self.val = val\n",
    "        self.left = left\n",
    "        self.right = right\n",
    "\n",
    "    @staticmethod\n",
    "    def from_list(values: list[Optional[int]]) -> Optional[TreeNode]:\n",
    "        \"\"\"\n",
    "        Constructs a binary tree from a list of values using level-order traversal (BFS).\n",
    "\n",
    "        Parameters:\n",
    "        - values (list): The list of values to construct the binary tree from.\n",
    "\n",
    "        Returns:\n",
    "        TreeNode: The root of the constructed binary tree.\n",
    "        \"\"\"\n",
    "        if not values:\n",
    "            return None\n",
    "\n",
    "        root = TreeNode(val=values[0])\n",
    "        queue = deque([root])\n",
    "        i = 1\n",
    "\n",
    "        while queue and i < len(values):\n",
    "            current_node = queue.popleft()\n",
    "\n",
    "            if values[i] is not None:\n",
    "                current_node.left = TreeNode(val=values[i])\n",
    "                queue.append(current_node.left)\n",
    "            i += 1\n",
    "\n",
    "            if i < len(values) and values[i] is not None:\n",
    "                current_node.right = TreeNode(val=values[i])\n",
    "                queue.append(current_node.right)\n",
    "            i += 1\n",
    "\n",
    "        return root\n",
    "\n",
    "    def print_level_order(self):\n",
    "        \"\"\"\n",
    "        Prints the level order traversal of the tree (BFS).\n",
    "\n",
    "        Returns:\n",
    "        list: A list of lists, where each sublist contains the values of the tree nodes\n",
    "              at that depth.\n",
    "        \"\"\"\n",
    "        queue = deque([(self, 0)])\n",
    "        result = []\n",
    "        current_level = []\n",
    "        level_number = 0\n",
    "\n",
    "        while queue:\n",
    "            current_node, node_level = queue.popleft()\n",
    "\n",
    "            if node_level > level_number:\n",
    "                result.append(current_level)\n",
    "                current_level = []\n",
    "                level_number = node_level\n",
    "\n",
    "            if current_node:\n",
    "                current_level.append(current_node.val)\n",
    "                queue.append((current_node.left, node_level + 1))\n",
    "                queue.append((current_node.right, node_level + 1))\n",
    "            else:\n",
    "                current_level.append(\"None\")\n",
    "\n",
    "        if current_level:\n",
    "            result.append(current_level)\n",
    "        \n",
    "        # Remove trailing \"None\" values\n",
    "        result = [level for level in result if any(elem != \"None\" for elem in level)]\n",
    "\n",
    "        return result"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Solution:\n",
    "    # divide and conquer - find root in postorder, then find left and right subtrees in inorder\n",
    "    def buildTree(self, inorder: list[int], postorder: list[int]) -> Optional[TreeNode]:\n",
    "        if not inorder or not postorder:\n",
    "            return None\n",
    "\n",
    "        # the last element in postorder is the root\n",
    "        root_val = postorder.pop()\n",
    "        root = TreeNode(root_val)\n",
    "\n",
    "        # find the index of the root in inorder\n",
    "        root_index = inorder.index(root_val)\n",
    "\n",
    "        # find the right subtree in inorder\n",
    "        root.right = self.buildTree(inorder[root_index + 1:], postorder)\n",
    "        root.left = self.buildTree(inorder[:root_index], postorder)\n",
    "\n",
    "        return root"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[[3], [9, 20], ['None', 'None', 15, 7]]"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "# Test case\n",
    "\n",
    "s = Solution()\n",
    "\n",
    "root = s.buildTree([9, 3, 15, 20, 7], [9, 15, 7, 20, 3])\n",
    "root.print_level_order()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The reason for constructing the right subtree first in the solution, especially in the context of using the `postorder` traversal, is rooted in the structure of the `postorder` array itself. Let's break down the reasoning:\n",
    "\n",
    "### Postorder Traversal\n",
    "\n",
    "In postorder traversal, a tree is traversed in the following order: left subtree, right subtree, and then the root node. This means that the last element in a postorder sequence is the current tree's root node. The elements before the root are divided into two groups: all the elements belonging to the left subtree come first, followed by the elements belonging to the right subtree.\n",
    "\n",
    "### Why Build the Right Subtree First?\n",
    "\n",
    "Given how we pop the last element of the `postorder` list to determine the root, it's crucial to manage the remaining elements in the `postorder` array correctly, as they represent the left and right subtrees. Since the elements for the right subtree appear right before the root element in the postorder sequence, and we are popping the root element from the end of the `postorder` list, the sequence of elements that remains in the `postorder` list naturally aligns with the expectation for building the right subtree first:\n",
    "\n",
    "- After popping the root, the end of the `postorder` array directly aligns with the right subtree elements.\n",
    "- By recursively calling the function to build the right subtree first, we effectively consume the right subtree elements from the `postorder` list, which is crucial because the size of the right subtree dictates how many elements we should take from the end of the `postorder` list.\n",
    "- Once the right subtree is built and those elements are consumed, the `postorder` list now ends with the elements belonging to the left subtree, correctly positioning us to build the left subtree in the next recursive call.\n",
    "\n",
    "### The Recursive Calls\n",
    "\n",
    "The recursive construction is done in two parts, each time slicing the `inorder` array to pass only the relevant portion for the subtree being constructed:\n",
    "\n",
    "- `root.right = buildTree(inorder[root_index + 1:], postorder)`: For the right subtree, we use the part of the `inorder` array to the right of the root's position. This is because, in the inorder sequence, all elements to the right of the root belong to the right subtree.\n",
    "- `root.left = buildTree(inorder[:root_index], postorder)`: For the left subtree, we use the part of the `inorder` array to the left of the root's position, consistent with how inorder traversal works (elements to the left of the root belong to the left subtree).\n",
    "\n",
    "This methodical approach, considering the nature of `postorder` and `inorder` traversals, ensures the binary tree is reconstructed accurately from the given traversal sequences."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Let's use the example given in the question to illustrate why we build the right subtree first when reconstructing a binary tree from inorder and postorder traversals.\n",
    "\n",
    "### Example:\n",
    "\n",
    "- **Inorder traversal**: `[9,3,15,20,7]`\n",
    "- **Postorder traversal**: `[9,15,7,20,3]`\n",
    "\n",
    "### Understanding the Traversal Orders:\n",
    "\n",
    "- **Inorder**: Left subtree → Node → Right subtree\n",
    "- **Postorder**: Left subtree → Right subtree → Node\n",
    "\n",
    "### Steps to Reconstruct the Tree:\n",
    "\n",
    "#### 1. Identify the Root:\n",
    "\n",
    "- The last item in the postorder list is the root of the tree. In this case, `3` is the root.\n",
    "\n",
    "#### 2. Locate the Root in the Inorder List:\n",
    "\n",
    "- Find `3` in the inorder list. It divides the tree into the left and right subtrees.\n",
    "- **Inorder left subtree**: `[9]`\n",
    "- **Inorder right subtree**: `[15,20,7]`\n",
    "\n",
    "#### 3. Build Right Subtree First:\n",
    "\n",
    "- Why right first? Because when we pop `3` (the root) from the postorder, we're left with `[9,15,7,20]`, which ends with the right subtree's nodes in postorder.\n",
    "\n",
    "#### Example Breakdown for Right Subtree:\n",
    "\n",
    "- **Remaining Postorder**: `[9,15,7,20]` (after removing `3`)\n",
    "- **Right Subtree Inorder**: `[15,20,7]`\n",
    "- The last element in the remaining postorder (`20`) is the root of the right subtree.\n",
    "- In the right subtree's inorder (`[15,20,7]`), `20` divides it into `[15]` (left) and `[7]` (right).\n",
    "\n",
    "By building the right subtree first, we utilize the postorder's natural ordering. After determining the right subtree, the postorder array directly corresponds to what's left to process, fitting perfectly for constructing the left subtree.\n",
    "\n",
    "#### Constructing the Right Subtree:\n",
    "\n",
    "- **Root**: `20`\n",
    "- **Left Child**: Found by locating `20` in the inorder array and taking everything to the left of it, which is `[15]`. In postorder, `15` is immediately before `20`, so it matches.\n",
    "- **Right Child**: Everything to the right of `20` in the inorder array, which is `[7]`. Since `7` is before `20` in postorder (after removing elements related to the left child and the root), it aligns with the construction order.\n",
    "\n",
    "#### Constructing the Left Subtree:\n",
    "\n",
    "- Now, we're left with `[9]` in both inorder and postorder for the left subtree, clearly indicating that `9` is a leaf node on the left.\n",
    "\n",
    "### Visual Representation:\n",
    "\n",
    "After building both subtrees, we have the tree:\n",
    "\n",
    "```\n",
    "     3\n",
    "    / \\\n",
    "   9  20\n",
    "     /  \\\n",
    "    15   7\n",
    "```\n",
    "\n",
    "- The process starts from the root found at the end of postorder.\n",
    "- Building the right subtree first leverages the postorder array's natural order after the root is removed.\n",
    "- This approach, rooted in the structure of inorder and postorder traversals, ensures that we correctly reconstruct the tree by systematically consuming the postorder array to reflect the true structure of the tree."
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
